// DISPLAY CODE

// File contains code from Jumpcore; notice applies to that code only:
/* Copyright (C) 2008-2009 Andrew McClure
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// File contains code from Chipmunk "MoonBuggy" demo; notice applies to that code only:
/* Copyright (c) 2007 Scott Lembcke
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "kludge.h"

#include "chipmunk.h"
#include "lodepng.h"
#include "internalfile.h"
#include "FTGLTextureFont.h"
#include <queue>
#include <string>
#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include "tinyxml.h"
#include "program.h"
#include "color.h"

// Display state: You probably want to keep all of this:

#define DRAW_DEBUG 1

FTFont* uiFont = NULL;
GLuint ftex;
float uiFontHeight = 0;

// Puts screen in orthographic projection where the screen top is 1.0, bottom is -1.0, left is -aspect, right is aspect
// Note: This is used by ControlBase, so if you mess with it that may stop working.
// If you need a different orthographic projection I recommend duplicating this function rather than altering it.
void goOrtho() {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0f, 1.0f, -1.0, 1.0, -1.0, 5.0);
	glScalef(1.0*aspect, 1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
}

void orthoText() { // Puts screen in a 1:1 pixel:point relationship. Needed by drawText.
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-surfacew/2, surfacew/2, -surfaceh/2, surfaceh/2, -1.0, 5.0);
	glMatrixMode(GL_MODELVIEW);
}

void goPerspective() { // An example projective ... projection matrix. Not used anywhere right now
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, 1/aspect, 0.5, 5);
	glMatrixMode(GL_MODELVIEW);
}

// Used by font stuff:

static float tempTexture[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0};

float textWidth(string str) {
	float llx, lly, llz, urx, ury, urz;
	uiFont->BBox(str.c_str(), llx, lly, llz, urx, ury, urz);
	return (urx - llx)/(surfaceh/2);	
}
float textHeight(string s = string()) {
	return uiFontHeight/(surfaceh/2);	
}
float centerOff(const char *str) {
	float llx, lly, llz, urx, ury, urz;
	uiFont->BBox(str, llx, lly, llz, urx, ury, urz);
	return (urx - llx)/2;
}
void initFont(int size) {
	uiFont->FaceSize(size);
	uiFontHeight = uiFont->LineHeight();
	
	glGenTextures(1, &ftex);
    glBindTexture(GL_TEXTURE_2D, ftex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 4, 4, 0, GL_RGB, GL_FLOAT, tempTexture);
}

// x and y are in range -1..1, -aspect..aspect, rot is in range 0..360
void drawText(string text, double x, double y, double rot = 0, bool xcenter = true, bool ycenter = true) {
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(-surfacew/2, surfacew/2, -surfaceh/2, surfaceh/2, -1.0, 5.0);
	glMatrixMode(GL_MODELVIEW);
	
	glPushMatrix();	
	glLoadIdentity();
	glColor3f(0.0,0.0,0.0);	
	glTranslatef(x*surfaceh/2, y*surfaceh/2, 0);
	if (rot)
		glRotatef(rot,0,0,1);
	glTranslatef(xcenter?-centerOff(text.c_str()):0, ycenter?-uiFontHeight/2:0, 0);
		
	glEnable( GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ftex);
	uiFont->Render(text.c_str());
	glDisable( GL_TEXTURE_2D); 	

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void drawWord(cpBody *word) { // Used by typewriter demo... delete this
	wordinfo *info = (wordinfo *)word->data;
	drawText(info->text, word->p.x, word->p.y, word->a/M_PI*180, info->center, info->center);
}

// Used by DRAW_DEBUG
void drawCollisions(void *ptr, void *data)
{
	cpArbiter *arb = (cpArbiter*)ptr;
	int i;
	for(i=0; i<arb->numContacts; i++){
		cpVect v = arb->contacts[i].p;
		glVertex2f(v.x, v.y);
	}
}

// The default display() iterates over each shape in the hash and calls a drawObject() on each one.
// This forwards to one of the following methods, depending on what kind of shape it is:
void drawCircle(cpFloat x, cpFloat y, cpFloat r, cpFloat a)
{
#if DRAW_DEBUG
	const int segs = 15;
	const cpFloat coef = 2.0*M_PI/(cpFloat)segs;
	
	int n;
	
	glColor3f(0.5, 0.5, 0.5);
	glBegin(GL_LINE_LOOP); {
		for(n = 0; n < segs; n++){
			cpFloat rads = n*coef;
			glVertex2f(r*cos(rads + a) + x, r*sin(rads + a) + y);
		}
		glVertex2f(x,y);
	} glEnd();
#endif
}

void drawCircleShape(cpShape *shape)
{
#if DRAW_DEBUG
	cpBody *body = shape->body;
	cpCircleShape *circle = (cpCircleShape *)shape;
	cpVect c = cpvadd(body->p, cpvrotate(circle->c, body->rot));
	drawCircle(c.x, c.y, circle->r, body->a);
#endif
}

void drawSegmentShape(cpShape *shape)
{
#if DRAW_DEBUG
	cpBody *body = shape->body;
	cpSegmentShape *seg = (cpSegmentShape *)shape;
	cpVect a = cpvadd(body->p, cpvrotate(seg->a, body->rot));
	cpVect b = cpvadd(body->p, cpvrotate(seg->b, body->rot));
	
	glBegin(GL_LINES); 
	glColor3f(0.5, 0.5, 0.5);{
		glVertex2f(a.x, a.y);
		glVertex2f(b.x, b.y);
	} glEnd();
#endif
}

void drawPolyShape(cpShape *shape, void *_data)
{
	cpBody *body = shape->body;
	
	switch(shape->collision_type) {
		case C_WORD:
			drawWord(body);
			break;
		default:break;
	}
	
#if DRAW_DEBUG
	if (optDebug) {
		cpPolyShape *poly = (cpPolyShape *)shape;
		int num = poly->numVerts;
		cpVect *verts = poly->verts;
		int i;
		glBegin(GL_LINE_LOOP);
		glColor3f(0.5f,0.5f,0.5f);
		for(i=0; i<num; i++){
			cpVect v = cpvadd(body->p, cpvrotate(verts[i], body->rot));
			glVertex3f(v.x, v.y, 0);
		} glEnd();
	}
#endif
}

void drawObject(void *ptr, void *data)
{
	cpShape *shape = (cpShape*)ptr;
	switch(shape->klass->type){
		case CP_CIRCLE_SHAPE:
			drawCircleShape(shape);
			break;
		case CP_SEGMENT_SHAPE:
			drawSegmentShape(shape);
			break;
		case CP_POLY_SHAPE:
			drawPolyShape(shape, data);
			break;
		default:
			ERR("Bad enumeration in drawObject().\n");
	}
}

// Callbacks follow:

// This is called once each time the display surface is initialized. It's a good place to do things like initialize
// fonts and textures. (Note it could be called more than once if the window size ever changes.)
void display_init() {
	if (uiFont)
		delete uiFont;
	
	char fontfile[FILENAMESIZE];
	internalPath(fontfile, "DroidSans.ttf");
	uiFont = new FTGLTextureFont( fontfile);
	
	if ( !uiFont ){
		REALERR("Couldn't open font file: %s\n", fontfile);
		Quit(1);
	}
	
	if ( uiFont->Error() ){
		REALERR("Couldn't open font file (error %d): %s\n", (int)uiFont->Error(), fontfile);
		Quit(1);
	}
	
	float fontSize = 18;
	fontSize *= (surfaceh/800.0); // Scale the font size so that it will be 18 point on a 1024x800 screen.
	
	initFont(fontSize);
}

// This is called when it is time to draw a new frame.
void display(void)
{	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	glTranslatef(0, 0, -1); // So we can draw at 0,0,0

	goOrtho();
	
	// Do we really want to be using these?
	for(int c = spaces.size()-1; c >= 0; c--) { // FIXME: Forward or backward?
		spaceinfo *s = &(spaces[c]);
		
		glDisable(GL_DEPTH_TEST); // You probably do not want to disable depth test.
		glPushMatrix();
		
		cpSpaceHashEach(s->space->staticShapes, &drawObject, s);
		
		cpSpaceHashEach(s->space->activeShapes, &drawObject, s);
		
		glBegin(GL_POINTS); { // DELETE ME?
			if (draggingMe) {
				glColor3f(0.0, 0.0, 1.0);
				glVertex2f(lastClick.x, lastClick.y);
			}
		} glEnd();
		
		// Debug code-- feel free to remove completely.
#if DRAW_DEBUG
		if (optDebug) {
			// Draw positions of bodies.
			cpArray *bodies = s->space->bodies;
			
			glBegin(GL_POINTS); {
				glColor3f(0.0, 0.0, 1.0);
				for(int i=0; i<bodies->num; i++){
					cpBody *body = (cpBody*)bodies->arr[i];
					glVertex2f(body->p.x, body->p.y);
				}
				glColor3f(1.0, 0.0, 0.0);
				cpArrayEach(s->space->arbiters, &drawCollisions, NULL);
			} glEnd();
			
#if AXIS_DEBUG
			glBegin(GL_LINES); { 
				for(int i=0; i<bodies->num; i++){
					cpBody *body = (cpBody*)bodies->arr[i];
					glColor3f(1.0, 0.0, 1.0);
					glVertex2f(body->p.x, body->p.y);
					glVertex2f(body->p.x+body->rot.x*36, body->p.y+body->rot.y*36);
				}
			} glEnd();
#endif		
		}
#endif
		
		glPopMatrix();
		glEnable(GL_DEPTH_TEST);
	}
	
	// Draw word-in-progress, cursor
	drawWord(nextWord);
	if (optCursor && !( (ticks-lastLetterAt)/TPF/3 % 2 )) {
		float x = nextWord->p.x + textWidth(((wordinfo *)nextWord->data)->text);
		float y = nextWord->p.y + 2.0/surfaceh;
		drawText("|", x, y, 0, false, false);
	}

	
	{ // Draw "interface"
		goOrtho(); // Will be redundant if nothing since the beginning of the function has messed with the projection matrix.
		glLoadIdentity();
		glDisable(GL_DEPTH_TEST);

		cpSpaceHashEach(workingUiSpace()->staticShapes, &drawButton, NULL);

		glEnable(GL_DEPTH_TEST);
		goPerspective();
	}
	
	
	SDL_GL_SwapBuffers();
	ticks++;
	
	program_update();
}

// Rewrite this if you want to change how ControlBase items appear.
void drawButton(void *ptr, void *data)
{
	cpPolyShape *poly = (cpPolyShape *)ptr;
	ControlBase *control = (ControlBase *)poly->shape.data;
	int num = poly->numVerts;
	cpVect *verts = poly->verts;
	
	if (!control) return;
	
	// Note we do no body-specific translation because we assume this is the UI layer.
	if (control->bg) {
		glBegin(GL_QUADS);
		if (control == KeyboardControl::focus || control->highlighted)
			glColor4f(0.5,0.5,1.0,0.5);	
		else
			glColor4f(0.75,0.75,0.75,0.75); 
		for(int i=0; i<num; i++){
			cpVect v = verts[i];
			glVertex3f(v.x, v.y, 0);
		}
		glEnd();
	}
	
	bool floatOff = !control->text.empty() && control->img; // "This button has both text and an icon"
	const double border = button_height/18;
	const double slicesize = button_height - border*2; // I have no idea
	
	if (!control->text.empty()) { // Draw button text -- a little complicated because it recognizes newlines
		int count = 1, index = 0;
		string::size_type upto = 0, from = 0;
		while (string::npos != (upto = control->text.find('\n', upto+1))) // Assumes the first character is never a \n.
			count++;
		
		upto = 0;
		from = 0;
		do {
			upto = control->text.find('\n', upto+1);
			string sub = control->text.substr(from, string::npos == upto ? control->text.size() : upto-from);
			
			glColor3f(0.0,0.0,0.0);	
			
			cpVect at = control->p;
			at.y += (count/2.0-index-1)*textHeight();
			at.y -= uiFont->Descender()/surfaceh;
			if (floatOff) 
				at.x += slicesize/2;
			drawText(sub, at.x, at.y, 0, true, false);
			
			index++;
			from = upto + 1;
		} while (upto != string::npos);
	} 

	if (control->img) {
		glPushMatrix();

		glTranslatef(control->p.x, control->p.y, 0);
		if (floatOff)
			glTranslatef(-button_width/2 + slicesize/2 + border, 0, 0);
		glColor3f(0,0,0);
		glEnable( GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, control->img->texture);
		glBegin(GL_QUADS);
		glTexCoord2d(0,0); glVertex3f(-slicesize/2, slicesize/2, 0);
		glTexCoord2d(0,1); glVertex3f(-slicesize/2, -slicesize/2, 0);
		glTexCoord2d(1,1); glVertex3f(slicesize/2, -slicesize/2, 0);
		glTexCoord2d(1,0); glVertex3f(slicesize/2, slicesize/2, 0);
		glEnd();
		glDisable( GL_TEXTURE_2D); 		
		glPopMatrix();
	}
}

// A maybe-overcomplicated function for displaying a fatal error (given as "why"). All event handling will halt and a
// window will go up explaining the error; on the next keypress after this, the program will quit. This uses the global
// "uiFont" font (but is not dependent on drawText()). See also FileBombBox
void BombBox(string why) {
	glClearColor(1.0,1.0,1.0,0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	orthoText();
	
	float leftMargin = -centerOff("(Press a key or gamepad button to set the highlighted control)");
	int lineCount = -1;
	
	glColor3f(0.0, 0.0, 0.0);
	
	glPushMatrix();
	glTranslatef(leftMargin-72, 200, 0);
	glRotatef(180,0,1,0);
	//	glScalef(1.5, 1.5, 1.0);
	
	glPopMatrix();
	
	glEnable( GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, ftex);
	glPushMatrix();
	
	glTranslatef(leftMargin, 200, 0);
	
	glPushMatrix();
	uiFont->Render("FATAL ERROR");
	glPopMatrix();
	glTranslatef(0, -uiFontHeight, 0);
	glTranslatef(0, -uiFontHeight, 0);
	while (!why.empty()) { // Wraps text. I can't believe I have to do simple stuff like this myself..
		char filename2[FILENAMESIZE+1]; memset(filename2, 0, FILENAMESIZE+1);
		int size = why.size(); if (size>FILENAMESIZE) size = FILENAMESIZE;
		int c = 0;
		for(; c < size; c++) {
			char next = why[c];
			if ('\n' == next) { 
				if (c > 0 && c+1 == why.size()) // Special case: '\n' comes at end of string
					c--;
				break;
			}
			filename2[c] = next;
			if (c > 0 && -centerOff(filename2) < leftMargin-36) {
				filename2[c] = '\0';
				c--;
				break;
			}
		}
		glPushMatrix();
		uiFont->Render(filename2);
		glPopMatrix();
		glTranslatef(0, -uiFontHeight, 0);
		lineCount++;
		
		why.replace(0, c+1, "");
	}
	glPushMatrix();
	uiFont->Render("The program must shut down.");
	glPopMatrix();
	glTranslatef(0, -uiFontHeight, 0);
	glTranslatef(0, -uiFontHeight, 0);
	glPushMatrix();
	uiFont->Render("Press any key to continue.");
	glPopMatrix();
	
	glDisable( GL_TEXTURE_2D); 
	glPopMatrix();
	
	glBegin(GL_LINE_LOOP);
	glVertex2f( leftMargin-72*2, 200+72);
	glVertex2f(-leftMargin+72*2, 200+72);
	glVertex2f(-leftMargin+72*2, 100-72 - lineCount*uiFontHeight);
	glVertex2f( leftMargin-72*2, 100-72 - lineCount*uiFontHeight);
	glEnd();
	
	SDL_GL_SwapBuffers();		
	
	while(1) { SDL_Event event;
		while ( SDL_PollEvent(&event) ) {
			//		ERR("OK event %d (vs %d) ... axis %d value %d\n", (int) event.type, (int)SDL_JOYAXISMOTION, (int)event.jaxis.axis, (int)event.jaxis.value);
			if ( event.type == SDL_QUIT || event.type == SDL_KEYDOWN) {
				Quit(1);
			}
		}
	}
}

// Calls BombBox with a message stating "filename" could not be fou
void FileBombBox(string filename) {
	string because;
	if (filename.empty())
		because = "An internal file could not be opened.";
	else
		because = "Could not open file:\n";
	BombBox(because + filename + "\n");
}

// Audio state:

#define NO_AUDIO 0

FILE *audiolog = NULL; // If you find yourself needing to debug your audio, fopen() this somewhere and all audio will start getting copied to this file as 16-bit pcm

int tapping = 0; // Used by typewriter demo
double lastWhiteNoiseValue = 0; // Used by typewriter demo

// Audio code:

void audio_callback(void *userdata, uint8_t *stream, int len) {
	int16_t *samples = (int16_t *)stream;
	int slen = len / sizeof(int16_t);

	for(int c = 0; c < slen; c++) {
		double value = 0;
		
		// This block here is part of the typewriter demo, clear and replace with your own code:
		if (tapping > 0) { // If a keypress has recently happened
			value = (double)random()/RANDOM_MAX * ((double)tapping/TAPFULL * 0.5); // Play some white noise, which loses volume with time.
			value = value/2 + lastWhiteNoiseValue/2; // It'll sound harsh, so smooth it out by averaging it with the last white noise value we calculated.
			lastWhiteNoiseValue = value; // And save the old white noise value for next time.
			tapping--;
		}
		
#if NO_AUDIO
		value = 0;
#endif
		samples[c] = value*SHRT_MAX;
	}
	if (audiolog) { // For debugging audio
		fwrite(stream, sizeof(short), slen, audiolog);
	}
}