/*
 *  program.h
 *  Jumpcore
 *
 *  Created by Andrew McClure on 4/27/08.
 *
 */

#include "kludge.h"
#include "controls.h"
#include <math.h>
#include <list>

#define NAME_OF_THIS_PROGRAM "Jumpcore"

// The following handful of lines are the only ones you can't erase. Feel free to redefine "FPS".

#define FPS 80.0
#define TPF ((int)(1000/FPS))
extern int ticks;
extern int sinceLastFrame;
extern double aspect;
extern bool gotJoystick;
extern bool paused;
extern int surfacew, surfaceh;
cpVect screenToGL(int screenx, int screeny, double desiredZ);

extern cpSpace *uiSpace;
inline cpSpace *workingUiSpace() { return uiSpace; }

void recreate_surface(bool in_a_window);

// CALLBACKS

void BackOut();
void AboutToQuit();
void Quit(int code = 0);
void BombBox(string why = string());
void FileBombBox(string filename = string());

void program_init(void);
void program_update(void);
void program_eventkey(SDL_Event &event);
void program_eventjoy(SDL_Event &event);
void program_eventmouse(SDL_Event &event);

void display(void);
void display_init();
void drawButton(void *ptr, void *data);
void audio_callback(void *userdata, uint8_t *stream, int len);

// Interface / ControlBase

extern vector<ContainerBase *> columns;
void program_interface();
extern bool wantClearUi;
enum WantNewScreen { // TODO: DELETE ?!
	WNothing = 0,
	WOptions
};
extern WantNewScreen want; 
extern bool &optWindow;
inline int iabs(int i) { return i < 0 ? -i : i; } // Yeah, I dunno.

void goOrtho();

struct spaceinfo {
	double depth;
	cpSpace *space;
	cpBody *staticBody;

	spaceinfo() {
		depth = 0;
		space = NULL;
		staticBody = NULL;
	}
};

extern vector<spaceinfo> spaces;

inline bool okayToDraw(spaceinfo *) { return true; }

float textWidth(string str);
float textHeight(string str);

// Everything from here until the end of the file is specific to the "typewriter" demo and can/should be removed

#define TAPFULL 5000
extern int tapping;

#define C_WORD 0x9999
struct wordinfo {
	string text;
	bool center;
	wordinfo(bool _center = true) : center(_center) {}
};
extern bool optDebug, optCursor;
extern cpBody *nextWord;
extern cpVect wordOrigin;

cpBody *allocWord();

extern cpVect lastClick; // unnecessary?
extern cpShape *draggingMe;
extern cpVect draggingMeInto;
extern unsigned int lastLetterAt;